(*muXQ core language implementation *)
open Xml;;
open Printer;;

type tvar = string
type var = string
type lab = string

type exp = String of string
  | Empty
  | NodeTest of exp * lab
  | Child of tvar
  | Tvar of tvar
  | Var of var
  | Seq of exp * exp
  | Let of var * exp * exp
  | For of tvar * exp * exp
  | Tree of lab * exp
  | Eq of exp * exp
;;


type tree = Xml.xml
and forest = tree list;;

type env = {tenv : tvar -> tree; env : var -> forest};;

let childr t = 
  match t with 
    PCData  _ -> None
  | Element(l,_,f) -> Some f
;;


let rec match_label f l = 
  match f with
    [] -> []
  | t::ts -> match_label_t t l @ match_label ts l 
and match_label_t t l = 
  match t with 
    PCData _ -> []
  | Element(l',atts,f) -> if l = l' then [Element(l,atts,f)] else []
;;


let init_env v = {env=(fun y -> if y = "$doc" then v else raise Not_found);
		  tenv = (fun tvc -> raise Not_found)}
;;

let empty_env = {env=(fun vc -> raise Not_found);
		 tenv = (fun tvc -> raise Not_found)}
;;


let update_env env x v = 
  {env=(fun y -> if x = y then v else env.env y);
    tenv = env.tenv}
;;


let update_tenv env x t = 
  {env=env.env;
   tenv = (fun y -> if x = y then t else env.tenv y)}
;;


let rec eval env exp = 
  match exp with 
    String s -> [PCData s]
  | Empty -> []
  | Seq (e1,e2) -> eval env e1 @ eval env e2
  | Tree (l,e) -> [Element (l,[],eval env e)]
  | NodeTest (e,l) -> match_label (eval env e) l
  | Child tv -> (
      match childr(env.tenv tv) with
	Some f -> f
     )
  | Tvar (tv) -> [env.tenv tv]
  | Var (v) -> env.env v
  | Let (x,e1,e2) -> eval (update_env env x (eval env e1)) e2
  | For (tx,e1,e2) -> 
      List.concat 
	(List.map (fun t -> eval (update_tenv env tx t) e2)  
	   (eval env e1))
  | Eq (e1,e2) -> (
      match eval env e1, eval env e2 with
	[v1], [v2] -> if v1 = v2 then [Element("true",[],[])] else []
     )
;;


let rec pp exp = 
  match exp with 
    String(s) -> quotes (text s)
  | Empty -> text "()"
  | NodeTest (e,l) -> pp e <:> cons <:> text l
  | Child(tv) -> text tv <:> slash <:> text "child"
  | Tvar(tv) -> text tv
  | Var(v) -> text v
  | Seq(e1,e2) -> paren (pp e1 <:> comma <:> pp e2)
  | Let(v,e1,e2) -> 
      text "let" <+> text v <+> assign <+> pp e1 <+> text "in" <:>
      newline <:> pp e2
  | For (tv,e1,e2) -> 
      text "for" <+> text tv <+> text "in"  <+> pp e1 <+> text "return" <:> 
      newline <:> 
      pp e2
  | Tree(l,e) -> text l <:> bracket(pp e)
  | Eq(e1,e2) -> paren (pp e1 <+> eq <+> pp e2)
