(* Lux: tiny update language *)
open Printer;;
open Muxq;;
open Xml;;

type stmt = Skip
  | IfThenElseStmt of exp * stmt * stmt
  | LetStmt of var * exp * stmt
  | SeqStmt of stmt * stmt
  | Replace of exp
  | Snapshot of var * stmt
  | Insert of exp
  | Delete
  | LabelTest of lab * stmt
  | StringTest of stmt
  | Left of stmt
  | Right of stmt
  | Children of stmt
  | Iterate of stmt
;;


let rec exec env f stmt = 
  match stmt with
    Skip -> f
  | IfThenElseStmt(c,s1,s2) -> (
      match (eval env c)
      with [] -> exec env f s2
      | _ -> exec env f s1
    )
  | LetStmt(x,e1,s2) -> 
      exec (update_env env x (eval env e1)) f s2
  | SeqStmt(s1,s2) -> exec env (exec env f s1) s2
  | Replace(e) -> eval env e
  | Snapshot(x,s) -> exec (update_env env x f) f s
  | Insert(e) -> (
      match f with
	[] -> eval env e
     )
  | Delete -> []
  | LabelTest (l,u) -> (
      match f with
	[Element(l',_,_)] -> if l' = l then exec env f u else f
      | [t] -> [t]
     )
  | StringTest(u) -> (
      match f with
	[PCData _] -> exec env f u
      | [t] -> [t]
     )
  | Left(u) -> (exec env [] u) @ f
  | Right(u) -> f @ (exec env [] u) 
  | Children(u) -> (
      match f with
	[Element(l,atts,f')] -> [Element(l,atts,exec env f' u)]
     )
  | Iterate(u) -> 
      List.concat (List.map (fun t -> exec env [t] u) f)
;;


let rec pp stmt = 
  match stmt with
    Skip -> text "skip"
  | IfThenElseStmt(c,s1,s2) -> 
      text "if" <+> (indent 2 (Muxq.pp c)) <:> newline <:>
      text "then" <+> pp s1 <:> newline <:>
      text "else" <+> pp s2
  | LetStmt(v,e1,s2) ->  
      text "let" <+> text v <+> assign <+> (indent 2 (Muxq.pp e1)) <+> text "in" <:> 
      newline <:> 
      pp s2
  | SeqStmt(s1,s2) -> 
      pp s1 <:> semi <:> newline <:> 
      pp s2
  | Replace(e) -> text "replace" <+> (indent 2 (Muxq.pp e))
  | Snapshot(x,s) -> 
      text "snapshot" <+> text x <+> text "in" <:> 
      newline <:> 
      (indent 4 (pp s))
  | Insert(e) -> text "insert" <+> (indent 2 (Muxq.pp e))
  | Delete -> text "delete"
  | StringTest(u) -> text "string" <:> question <+> pp u
  | LabelTest (l,u) -> text l <:> question <+> pp u
  | Left(u) -> text "left" <:> braces (indent 2 (newline <:> pp u)<:> newline)
  | Right(u) -> text "right" <:> braces (indent 2 (newline <:> pp u)<:> newline)
  | Children(u) -> text "children" <:> braces (indent 2 (newline <:> pp u)<:> newline)
  | Iterate(u) -> text "iterate" <:> braces (indent 2 (newline <:> pp u) <:> newline)
;;


